# Copyright (C) 2023 University of Bordeaux, CyVi Group & Anish Koyamparambath
# This file is part of geopolrisk-py library.
#
# geopolrisk-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# geopolrisk-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with geopolrisk-py.  If not, see <https://www.gnu.org/licenses/>.

import unittest, os
from .defaults import *
from geopolrisk.assessment.utils import *
from geopolrisk.assessment.core import *
from geopolrisk.assessment.__init__ import instance

_wgi = instance.wgi
# testcase to test the gprs_comtrade function
Resources = random_sampler(ListofMetals, 5)
Countries = random_sampler(ListofCountries, 5)
Year = random_sampler(ListofYears, 5)
listofcountryname, listofresourcename = [], []
for i, n in enumerate(Resources):
    listofcountryname.append(ListofCountryName[ListofCountries.index(Countries[i])])
    listofresourcename.append(ListofMetalName[ListofMetals.index(n)])


class TestCoreFunctions(unittest.TestCase):
    def test_regioninsertion(self):
        NewRegion = {
            "West Europe": [
                "France",
                "Germany",
                "Italy",
                "Spain",
                "Portugal",
                "Belgium",
            ]
        }
        result = regions(NewRegion)
        from geopolrisk.assessment.__init__ import instance

        regionslist = instance.regionslist
        self.assertEqual(
            regionslist["West Europe"],
            NewRegion["West Europe"],
            msg="Region insertion failed",
        )
        self.assertIsNotNone(result, msg="Error in region creation!")

    def test_tradepathinsertion(self):
        result = settradepath("geopolrisk/tests/sample trade data.xlsx")
        self.assertIsNotNone(result, msg="Error in tradepath creation!")

    def test_worldtrade(self):
        regions()
        randomresource = random.choice(Resources)
        randomcountry = random.choice(Countries)
        randomyear = random.choice(Year)
        result = worldtrade(randomyear, randomcountry, randomresource)
        self.assertIsNotNone(result, msg="Error in worldtrade calculation!")

    def test_specifictrade(self):
        xxx = settradepath("geopolrisk/tests/sample trade data.xlsx")
        result = specifictrade(sheetname="test")
        self.assertIsNotNone(result, msg="Error in specific trade calculation!")

    def test_productiondata(self):
        regions()
        randomresource = random.choice(listofresourcename)
        randomcountry = random.choice(listofcountryname)
        result = ProductionData(randomresource, randomcountry)
        self.assertIsNotNone(result, msg="Error in production data calculation!")

    def test_weightedaverage(self):
        TradeData = [
            [
                40,
                56,
                76,
            ],
            [
                "Austria",
                "Belgium",
                "Brazil",
            ],
            [
                1723782,
                2155207,
                53399700,
            ],
        ]
        PIData = _wgi
        Year = "2015"
        recyclingrate = 0
        scenario = 0
        numerator, tradetotal = weightedtrade(
            Year,
            TradeData=TradeData,
            PIData=PIData,
            scenario=scenario,
            recyclingrate=recyclingrate,
        )
        res_num, res_total = [31532151.035755564, 57278689]
        self.assertEqual(
            [numerator, tradetotal],
            [res_num, res_total],
            msg="Weighted average calculation failed!",
        )
        self.assertIsNotNone(numerator, msg="Weighted average calculation failed!")

    def test_GeoPolRisk(self):
        ProductionData = [
            [   0.289,
                0.304,
                0.287,
                0.258,
                0.213,
                0.222,
                0.218,
            ],
            [   23659.0,
                25000.0,
                25000.0,
                21000.0,
                20000.0,
                20000.0,
                20000.0,
            ],
            [   2014,
                2015,
                2016,
                2017,
                2018,
                2019,
                2020,
            ],
        ]
        WTAData = [31532151.035755564, 57278689]
        Year = "2015"
        AVGPrice = 2010.22
        result = GeoPolRisk(ProductionData, WTAData, Year, AVGPrice)
        theresult = [0.304, 0.3832359438269072, 0.11650372692337978, 234.1981219359165]
        self.assertEqual(result, theresult, msg = "The GeoPolRisk calculation failed!")

    def test_redistribution(self):
        TradeData = [
            [
                40,
                56,
                76,
            ],
            [
                "Austria",
                "Belgium",
                "Brazil",
            ],
            [
                1723782,
                2155207,
                53399700,
            ],
        ]
        PIData = _wgi
        Year = "2015"
        recyclingrate = 10
        scenario = 1
        numerator, tradetotal = weightedtrade(
            Year,
            TradeData=TradeData,
            PIData=PIData,
            scenario=scenario,
            recyclingrate=recyclingrate,
        )
        res_num, res_total = [31532151.035755564, 57278689]
        self.assertLessEqual(
            numerator,
            res_num,
            msg="Weighted average calculation failed!"
        )
        self.assertIsNotNone(numerator, msg="Weighted average calculation failed!")



# Copyright (C) 2023 University of Bordeaux, CyVi Group & Anish Koyamparambath
# This file is part of geopolrisk-py library.
#
# geopolrisk-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# geopolrisk-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with geopolrisk-py.  If not, see <https://www.gnu.org/licenses/>.

import pandas as pd, json, sqlite3
import comtradeapicall as ctac
from urllib.request import Request, urlopen
from .__init__ import instance, logging, execute_query

# Define Paths
tradepath = None
_production, _reporter = instance.production, instance.reporter
regionslist, _outputfile = instance.regionslist, instance.exportfile
_price = instance.price
db = _outputfile + "/" + instance.Output
# Extract list of all data
HS = _price.HS.to_list()
HS = [int(float(x)) for x in HS]
Resource = _price.Resource.to_list()
Country = _reporter.Country.to_list()
ISO = _reporter.ISO.to_list()
ISO = [int(x) for x in ISO]


def convertCodes(resource, country):
    # Verify direction
    def check_variables(A, B):
        if isinstance(A, list) and isinstance(B, list):
            if all(isinstance(element, str) for element in A) and all(
                isinstance(element, str) for element in B
            ):
                return "numeric"
            elif all(isinstance(element, int) for element in A) and all(
                isinstance(element, int) for element in B
            ):
                return "text"
            else:
                return None
        elif isinstance(A, str) and isinstance(B, str):
            return "numeric"
        elif isinstance(A, int) and isinstance(B, int):
            return "text"
        else:
            return None

    def return_variables(X, A, B):
        if isinstance(X, list):
            X_transform = []
            for i, n in enumerate(X):
                try:
                    idx = A.index(n)
                    X_transform.append(B[idx])
                except Exception as e:
                    if n in regionslist.keys():
                        logging.debug(f"Region transformation cannot be applied!")
                        X_transform.append(n)
                    else:
                        logging.debug(f"Failed to transform! {e}")
                        logging.debug(f"Transformation failed for {n}")
                        X_transform = None
            return X_transform
        else:
            try:
                idx = A.index(X)
                X_transform = B[idx]
            except Exception as e:
                if X in regionslist.keys():
                    logging.debug(f"Region transformation cannot be applied!")
                    X_transform = X
                else:
                    logging.debug(f"Failed to transform! {e}")
                    logging.debug(f"Transformation failed for {X}")
                    X_transform = None
            return X_transform

    direction = check_variables(resource, country)

    commodity, countryname, HSCode, ISOCode = None, None, None, None
    if direction.lower() == "numeric":
        ResourceCX = return_variables(resource, Resource, HS)
        CountryCX = return_variables(country, Country, ISO)
        commodity = resource
        countryname = country
        HSCode = ResourceCX
        ISOCode = CountryCX
    elif direction.lower() == "text":
        ResourceCX = return_variables(resource, HS, Resource)
        CountryCX = return_variables(country, ISO, Country)
        commodity = ResourceCX
        countryname = CountryCX
        HSCode = resource
        ISOCode = country
    else:
        logging.debug("Failed to transform!")
        logging.debug(direction)
        ResourceCX, CountryCX = resource, country
    return commodity, countryname, HSCode, ISOCode


def callapirequest(period, country, commoditycode):
    period = str(period)
    country = str(country)
    commoditycode = str(commoditycode)
    try:
        get = ctac.previewTarifflineData(
            typeCode="C",
            freqCode="A",
            clCode="HS",
            period=period,
            reporterCode=country,
            cmdCode=commoditycode,
            flowCode="M",
            partnerCode=None,
            partner2Code=None,
            customsCode=None,
            motCode=None,
            maxRecords=500,
            format_output="JSON",
            countOnly=None,
            includeDesc=True,
        )
    except Exception as e:
        logging.debug(f"Error while calling API! {e}")
        return None, None
    try:
        if get is not None or not isinstance(get, type(None)) or len(get) == 0:
            get["Qty"] = get.groupby(["partnerCode"])["qty"].transform(sum)
            get["CifValue"] = get.groupby(["partnerCode"])["cifvalue"].transform(sum)
            get = get.drop_duplicates(subset="partnerCode", keep="first")
            try:
                cifvalueToT = sum(get["CifValue"].to_list())
                totalQ = sum(get["Qty"].to_list())
                if totalQ == 0:
                    pricecif = 0
                else:
                    pricecif = cifvalueToT / totalQ
            except Exception as e:
                logging.debug(f"Error while extracting cifvalue! {e}")
                get, pricecif = None, None
        else:
            logging.debug(f"Problem with the new API call! {get}")
            get, pricecif = None, None
    except Exception as e:
        logging.debug(f"Error while extracting and combining data! {e}")
        get, pricecif = None, None
    return get, pricecif


def oldapirequest(period, country, commoditycode):
    period = str(period)
    country = str(country)
    commoditycode = str(commoditycode)
    url = f"""https://comtrade.un.org/api/get?max=50000&type=C&freq=A&px=HS&ps={period}&r={country}&p=all&cc={commoditycode}&rg=1&fmt=json"""
    logging.info(url)
    try:
        request = Request(url)
        response = urlopen(request)
        elevations = response.read()
    except Exception as e:
        logging.debug(url)
        logging.debug(f"Error while calling native API! {e}")
        return None

    try:
        data = json.loads(elevations)
        data = pd.json_normalize(data["dataset"])
    except Exception as e:
        logging.debug("Error while parsing JSON!")
        return None

    return data


def replace_values(list_to_replace, item_to_replace, item_to_replace_with):
    return [
        item_to_replace_with if item == item_to_replace else item
        for item in list_to_replace
    ]


def create_id(HS, ISO, Year):
    HS, ISO, Year = str(HS), str(ISO), str(Year)
    if len(HS) == 4:
        HSID = "xx" + HS
    elif len(HS) == 5:
        HSID = "x" + HS
    else:
        HSID = HS
    if len(ISO) == 2:
        ISOID = "x" + ISO
    elif len(ISO) == 1:
        ISOID = "xx" + ISO
    else:
        ISOID = ISO
    DBID = HSID + ISOID + Year
    return DBID


# Verify if the calculation is already stored in the database to avoid recalculation
def sqlverify(DBID):
    try:
        sql = f"SELECT * FROM recordData WHERE id = '{DBID}';"
        row = execute_query(
            f"SELECT * FROM recordData WHERE id = '{DBID}';",
            db_path=db,
        )

    except Exception as e:
        logging.debug(f"Database error in sqlverify - {e}, {sql}")
        row = None
    if not row:
        return False
    else:
        return True


# This function doesnt override the calculaiton
def recordData(
    Resource, Country, Year, RR, Scenario, GPRS, CF, HHI, WTA, LogFile, OuputList
):
    resource, country, HSCODE, ISO = convertCodes(Resource, Country)
    DBID = create_id(HSCODE, ISO, Year)
    if sqlverify(DBID) is True:
        sqlstatement = f"""SELECT 'recycling_rate' ,'scenario', 'geopolrisk'
        FROM recordData WHERE id = '{DBID}';
        """
        try:
            row = execute_query(sqlstatement, db_path=db)
        except Exception as e:
            logging.debug(f"Failed to execute statement {e} with {sqlstatement}")
        if str(row[0][0]) == str(RR) and str(row[0][1]) == str(Scenario):
            if str(row[0][2]) != str(GPRS):
                sqlstatement = f"""UPDATE recordData SET hhi =
                '{HHI}', wta = '{WTA}', geopolrisk = '{GPRS}', geopol_cf = '{CF}',
                log_ref = '{LogFile}'
                """
                try:
                    row = execute_query(sqlstatement, db_path=db)
                except Exception as e:
                    logging.debug(
                        f"Failed to execute statement {e} with {sqlstatement}"
                    )
                OuputList.append(
                    str(Year),
                    str(resource),
                    str(country),
                    str(RR),
                    str(Scenario),
                    str(GPRS),
                    str(CF),
                    str(HHI),
                    str(WTA),
                )
            else:
                logging.info("The database already exists with the data")
        else:
            sqlstatement = f""" INSERT INTO recordData (id, country, resource, year,
            recycling_rate, scenario, geopolrisk, hhi, wta, geopol_cf, resource_hscode, 
            iso, log_ref) VALUES ('{DBID}','{country}', '{resource}', '{Year}',
            '{RR}', '{Scenario}', '{GPRS}', '{HHI}', '{WTA}', '{CF}', '{HSCODE}',
            '{ISO}', '{LogFile}');
            """
            try:
                row = execute_query(sqlstatement, db_path=db)
            except Exception as e:
                logging.debug(f"Failed to execute statement {e} with {sqlstatement}")
            OuputList.append(
                [
                    str(Year),
                    str(resource),
                    str(country),
                    str(RR),
                    str(Scenario),
                    str(GPRS),
                    str(CF),
                    str(HHI),
                    str(WTA),
                ]
            )
    else:
        sqlstatement = f""" INSERT INTO recordData (id, country, resource, year,
            recycling_rate, scenario, geopolrisk, hhi, wta, geopol_cf, resource_hscode, 
            iso, log_ref) VALUES ('{DBID}','{country}', '{resource}', '{Year}',
            '{RR}', '{Scenario}', '{GPRS}', '{HHI}', '{WTA}', '{CF}', '{HSCODE}',
            '{ISO}', '{LogFile}');
            """
        try:
            row = execute_query(sqlstatement, db_path=db)
        except Exception as e:
            logging.debug(f"Failed to execute statement {e} with {sqlstatement}")
        OuputList.append(
            [
                str(Year),
                str(resource),
                str(country),
                str(RR),
                str(Scenario),
                str(GPRS),
                str(CF),
                str(HHI),
                str(WTA),
            ]
        )


"""Convert entire database to required format
**CHARACTERIZATION FACTORS
Refer to python json documentation for more information on types of
orientation required for the output.
"""

# Extract CFs
def generateCF(exportType="csv", orient=""):
    exportF = ["csv", "excel", "json"]
    if exportType in exportF:
        logging.debug("Exporting database in the format {}".format(exportType))
        CFType = exportType
    else:
        logging.debug(
            "Exporting format not supported {}. "
            "Using default format [csv]".format(exportType)
        )
        CFType = "csv"
    try:
        conn = sqlite3.connect(
            db,
            isolation_level=None,
            detect_types=sqlite3.PARSE_COLNAMES,
        )
        db_df = pd.read_sql_query("SELECT * FROM recorddata", conn)
        if CFType == "csv":
            db_df.to_csv(_outputfile + "/database.csv", index=False, encoding="utf-8")
        elif CFType == "excel":
            db_df.to_excel(
                _outputfile + "/database.xlsx", index=False, encoding="utf-8"
            )
        elif CFType == "json":
            db_df.to_json(_outputfile + "/database.json", orient=orient, index=False)
    except Exception as e:
        logging.debug(f"Error while exporting database! {e}")



# Copyright (C) 2023 University of Bordeaux, CyVi Group & Anish Koyamparambath
# This file is part of geopolrisk-py library.
#
# geopolrisk-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# geopolrisk-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with geopolrisk-py.  If not, see <https://www.gnu.org/licenses/>.

import unittest
from .defaults import *
from geopolrisk.assessment.utils import *
from geopolrisk.assessment.core import *

# testcase to test the gprs_comtrade function

#Extracting samples data from a list of resources, countries and years
Resources = random_sampler(ListofMetals, 5)
Countries = random_sampler(ListofCountries, 5)
Year = random_sampler(ListofYears, 5)
listofcountryname, listofresourcename = [], []
for i, n in enumerate(Resources):
    listofcountryname.append(ListofCountryName[ListofCountries.index(Countries[i])])
    listofresourcename.append(
        ListofMetalName[ListofMetals.index(n)]
    )


#Test class to test the functions in utils module
#Please note functions such as record data, extract data are excluded
class Testutilsfunctions(unittest.TestCase):
     
    #Test to verify the functionality of id creation 
    def test_verifyidcreation(self):
        Resources = [2606, 261710, 2524]
        Countries = [4, 36, 124]
        Year = [2016]
        result = [
            "xx2606xx42016",
            "xx2606x362016",
            "xx26061242016",
            "261710xx42016",
            "261710x362016",
            "2617101242016",
            "xx2524xx42016",
            "xx2524x362016",
            "xx25241242016",
        ]
        for i, n in enumerate(Resources):
            get = create_id(Resources[i], Countries[i], Year[0])
            self.assertIn(get, result, msg="Id creation failed!")

    #Test to verify the conversion of the codes
    def test_conversiontonumericcodes(self):
        __ignore, __ignore, hscode, iso = convertCodes(
            listofresourcename, listofcountryname
        )
        self.assertEqual(
            [hscode, iso],
            [Resources, Countries],
            msg="Code conversion failed!",
        )

    #Test to verify the reverse conversion of the codes
    def test_conversiontonames(self):
        resourcename, countryname, __ignore, __ignore = convertCodes(
            listofresourcename, listofcountryname
        )
        self.assertEqual(
            [resourcename, countryname],
            [listofresourcename, listofcountryname],
            msg="Code conversion failed!",
        )

    #Test to verify the functionality of the new Comtrade API
    def test_pythoncomtradeapi(self):
        randomresource = random.choice(Resources)
        randomcountry = random.choice(Countries)
        randomyear = random.choice(Year)
        TradeData, cifprice = callapirequest(randomyear, randomcountry, randomresource)
        self.assertIsNotNone(TradeData, msg="API call failed!")
        self.assertIsNotNone(cifprice, msg="API call failed!")

    #Test to verify the functionality of the old Comtrade API
    def test_oldcomtradeapi(self):
        randomresource = random.choice(Resources)
        randomcountry = random.choice(Countries)
        randomyear = random.choice(Year)
        TradeData = oldapirequest(randomyear, randomcountry, randomresource)
        self.assertIsNotNone(TradeData, msg="API call failed!")

    #Test to verify the sql verify function to check for existing data
    def test_verifydatabase(self):
        DBID = ["xx81061912014", "xx81077052003", "xx2846x562012", "2615102032017"]
        for i in DBID:
            result =  sqlverify(i)
            self.assertTrue(result, msg="SQL verification failed!")



# Copyright (C) 2023 University of Bordeaux, CyVi Group & Anish Koyamparambath
# This file is part of geopolrisk-py library.
#
# geopolrisk-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# geopolrisk-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with geopolrisk-py.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["test_case_1","test_case_2", "test_case_3", "defaults"] 
__author__ = "Anish Koyamparambath <CyVi- University of Bordeaux>"
__status__ = "beta"
__version__ = "2.5"
__data__ = "10 March 2023"



# Copyright (C) 2023 University of Bordeaux, CyVi Group & Anish Koyamparambath
# This file is part of geopolrisk-py library.
#
# geopolrisk-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# geopolrisk-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with geopolrisk-py.  If not, see <https://www.gnu.org/licenses/>.

ListofMetals = [
    2606,
    261710,
    2524,
    2511,
    8106,
    8107,
    2610,
    2701,
    810520,
    2603,
    7108,
    2504,
    2601,
    2607,
    283691,
    251910,
    2602,
    280540,
    2613,
    271111,
    2604,
    2709,
    2846,
    261610,
    2609,
    2611,
    2608,
    261510,
]

ListofYears = [
    2002,
    2003,
    2004,
    2005,
    2006,
    2007,
    2008,
    2009,
    2010,
    2011,
    2012,
    2013,
    2014,
    2015,
    2016,
    2017,
    2018,
    2019,
    2020,
]

ListofCountries = [
    36,
    124,
    97,
    251,
    276,
    392,
    826,
    842,
    32,
    40,
    50,
    56,
    191,
    203,
    208,
    246,
    300,
    381,
    348,
    372,
    442,
    528,
    579,
    616,
    724,
    642,
    620,
    703,
    705,
    752,
    757,
    76,
    64,
]

ListofMetalName = [
    "Aluminium",
    "Antimony",
    "Asbestos",
    "Barytes",
    "Bismuth",
    "Cadmium",
    "Chromium",
    "Coal",
    "Cobalt",
    "Copper",
    "Gold",
    "Graphite",
    "Iron",
    "Lead",
    "Lithium",
    "Magnesite",
    "Manganese",
    "Mercury",
    "Molybdenum",
    "NG",
    "Nickel",
    "Crude_Oil",
    "REE",
    "Silver",
    "Tin",
    "Tungsten",
    "Zinc",
    "Zirconium",
]
ListofCountryName = [
    "Australia",
    "Canada",
    "European Union",
    "France",
    "Germany",
    "Japan",
    "United Kingdom",
    "USA",
    "Argentina",
    "Austria",
    "Bangladesh",
    "Belgium",
    "Croatia",
    "Czechia",
    "Denmark",
    "Finland",
    "Greece",
    "Italy",
    "Hungary",
    "Ireland",
    "Luxembourg",
    "Netherlands",
    "Norway",
    "Poland",
    "Spain",
    "Romania",
    "Portugal",
    "Slovakia",
    "Slovenia",
    "Sweden",
    "Switzerland",
    "Brazil",
    "Bhutan",
]


import random


def random_sampler(A, n):
    return random.sample(A, n)



# Copyright (C) 2023 University of Bordeaux, CyVi Group & Anish Koyamparambath
# This file is part of geopolrisk-py library.
#
# geopolrisk-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# geopolrisk-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with geopolrisk-py.  If not, see <https://www.gnu.org/licenses/>.

import pandas as pd, json
from urllib.request import Request, urlopen
from pathlib import Path
from .__init__ import instance, logging, execute_query
from .utils import *


# Define class for storing the path information for trade data
class classPath:
    tradepath = None


classpath = classPath()

# Create instances of data required for calculation of the GeoPolRisk
try:
    # All instances are extracted from the library database
    _production, _reporter = instance.production, instance.reporter
    regionslist, _outputfile = instance.regionslist, instance.exportfile
    db = _outputfile + "/" + instance.Output
except Exception as e:
    logging.debug(f"Error with database files or init file {e}")


# Function to set the path of the trade data file for specific trade data calculation
# Provide the absolute path of the trade data file or relative to the current working directory
def settradepath(path):
    tradepath = path
    try:
        df = pd.read_excel(tradepath)
        classpath.tradepath = path
    except FileNotFoundError:
        logging.debug(f"File {tradepath} not found")
        tradepath = None
    except Exception as e:
        logging.debug(e)
        tradepath = None
    return tradepath


# Function to define new regions
# Provide a dictionary of the new regions
# The key should be the name of the negion
# The value should be the list of countries in the new region
# In case of using aggregate function from the main module,
# use the key as the country
def regions(*args):
    if len(args) != 0:
        for key, value in args[0].items():
            if type(key) is not str and type(value) is not list:
                logging.debug(
                    "Dictionary input to regions does not match required type."
                )
                return None
            Print_Error = [
                x
                for x in value
                if str(x) not in _reporter.Country.to_list()
                and str(x) not in _reporter["ISO"].astype(str).tolist()
            ]
            if len(Print_Error) != 0:
                logging.debug(
                    "Error in creating a region! "
                    "Following list of countries not"
                    " found in the ISO list {}. "
                    "Please conform with the ISO list or use"
                    " 3 digit ISO country codes.".format(Print_Error)
                )
                return None
            else:
                regionslist[key] = value
    # The function must be called before calling any other functions in the core
    # module. The following lines populate the region list with all the countries
    # in the world including EU defined in the init file.
    for i in _reporter.Country.to_list():
        if i in regionslist.keys():
            logging.debug("Country or region already exists cannot overwrite.")
        regionslist[i] = [i]
    return regionslist


# Extrade the trade data using the COMTRADE API
def worldtrade(
    year="2010",
    country="276",
    commodity="2602",
):
    try:
        # pricecif is the cif price of traded commodity
        # it is included in case of change of methodology or
        # unavailability of price data from USGS or LME
        data = oldapirequest(year, country, commodity)
    except Exception as e:
        logging.debug(f"Error with the comtrade API request: {e}")

    if data is None or data.shape[0] == 0:
        data, pricecif = callapirequest(year, country, commodity)
        if data is None or data.shape[0] == 0:
            logging.debug("API returned empty dataframe")
            TradeData = [None, None, None]
        else:
            if 0 in data.partnerCode.astype(int).to_list():
                err = data.partnerCode.to_list().index(0)
                data = data.drop(data.index[[err]])
                logging.info(
                    "Partner code 0 is found in the trade file. "
                    "Please check the trade file."
                )
            try:
                code = data.partnerCode.to_list()
                countries = data.partnerDesc.to_list()
                quantity = data.Qty.to_list()
                TradeData = [code, countries, quantity]
            except Exception as e:
                logging.debug(
                    f"The fetched dataframe from "
                    "the API does not have the required columns. {e}"
                )
                TradeData = [None, None, None]
    elif data is not None and data.shape[0] != 0:
        if data is not None and data.shape[0] != 0:
            Worldindex = data.ptCode.to_list().index(0)
            data = data.drop(data.index[[Worldindex]])
            code = data.ptCode.to_list()
            countries = data.ptTitle.to_list()
            quantity = data.TradeQuantity.to_list()
            TradeData = [code, countries, quantity]
        else:
            logging.debug("API returned empty dataframe")
            TradeData = [None, None, None]
    else:

        logging.info("API returned empty dataframe")
        TradeData = [None, None, None]
    return TradeData


# Extract trade data from the specific trade data file whose path is defined in
# the method earlier.
def specifictrade(sheetname=None):
    trade_path = classpath.tradepath

    # This function reads the trade file specific to an organization/company.
    # It is possible to read excel or csv file. Section 1 validates if the
    # path to the file is set using method 1 and reads the file into a dataframe.

    # 4.1 Section to validate the path to trade file
    if trade_path == None:
        logging.debug("Trade file path is not set.")
        return None
    else:

        # The file must be either in excel or csv file, failing which raises an
        # InputError from the warningsgprs module

        if Path(trade_path).suffix == ".xlsx" or Path(trade_path).suffix == ".xls":
            try:
                data = pd.read_excel(trade_path, sheet_name=sheetname)
            except Exception as e:
                logging.debug(f"Error while accessing the trade file {e}")
                return None
        elif Path(trade_path).suffix == ".csv":
            try:
                data = pd.read_csv(trade_path)
            except Exception as e:
                logging.debug(f"Error while accessing the trade file {e}")
                return None
        else:
            logging.debug(Path(trade_path).suffix)
            return None

        # The comtrade API results are categorized by imports to a region/country
        # per country and it includes the total imports to the region/country. The
        # code removes such imports to avoid erronous calculation.
        try:
            if data.shape[0] != 0:
                if 0 in data.ptCode.astype(int).to_list():
                    err = data.ptCode.to_list().index(0)
                    data = data.drop(data.index[[err]])
                    logging.info(
                        "Partner code 0 is found in the trade file. "
                        "Please check the trade file."
                    )
                try:
                    code = data.ptCode.to_list()
                    countries = data.ptTitle.to_list()
                    quantity = data.TradeQuantity.to_list()
                    TradeData = [code, countries, quantity]
                except Exception as e:
                    logging.debug(
                        f"The fetched dataframe from "
                        f"the API does not have the required columns. {e}"
                    )
                    TradeData = None
            else:
                logging.info("API returned empty dataframe")
                TradeData = None
            return TradeData
        except Exception as e:
            logging.debug(f"Error in the trade file {e}")
            return None


# Fetch the mineral resource ore production data from the library
def ProductionData(Resource, EconomicUnit):

    EconomicUnit = "EU" if EconomicUnit == "European Union" else EconomicUnit
    EconomicUnit = regionslist[EconomicUnit]
    try:
        prod = _production[Resource].fillna(0)
        Countries = prod.Country.to_list()
    except Exception as e:
        logging.debug(
            f"There was an error while acessing"
            f" the production data file with an exception as {e}"
        )
        return None

    # P2. Fetching the production quantity from 'prod' dataframe.
    try:
        Prod_Year = prod.columns.to_list()[1:-1]
        Prod_Year = [int(i) for i in Prod_Year]
        temp = [0] * len(Prod_Year)
        for i in EconomicUnit:
            if i in Countries:
                Prod_Qty = (
                    prod.loc[prod["Country"] == i]
                    .reset_index()
                    .loc[0, :]
                    .values.flatten()
                    .tolist()[2:-1]
                )
                Prod_Qty = replace_values(Prod_Qty, "^", 0)
                Prod_Qty = [float(i) for i in Prod_Qty]
                Prod_Qty = [sum(j) for j in zip(temp, Prod_Qty)]
                temp = Prod_Qty
            else:
                Prod_Qty = temp
    except Exception as e:
        logging.debug(f"Error while processing production data {e}")
        return None

    # P3. Calculating the HHI.
    HHI = []
    try:
        for i in Prod_Year:
            temp = prod[str(i)].values.tolist()
            temp = replace_values(temp, "^", 0)
            temp = [float(i) for i in temp]
            DeNom = sum(temp)
            Nom = sum([j[0] * j[1] for j in zip(temp, temp)])
            try:
                HHI.append(round(Nom / (DeNom * DeNom), 3))
            except Exception as e:
                HHI.append(0)
    except Exception as e:
        logging.debug(f"Error while calculating the HHI {e}")
        return None
    length = len(Prod_Year)
    if all(len(lst) == length for lst in [Prod_Qty, HHI]):
        return [HHI, Prod_Qty, Prod_Year]
    else:
        return None


# Weight the extracted trade data with the political instability indicator data.
# WGI is used as the default PI indicator that can be replaced with any other
# normalized indicator data.
def weightedtrade(year, TradeData=None, PIData=None, scenario=0, recyclingrate=0.00):
    if TradeData is None or PIData is None:
        logging.debug("Trade data or Indicator data returned empty!")
        return None
    elif TradeData is not None:
        code, quantity = TradeData[0], TradeData[2]
        reducedmass, totalreduce = 0, 0

        # 1.2 Section to calculate the numerator and trade total
        try:
            PIData = PIData.fillna(0)
            PIData.columns = PIData.columns.astype(str)
            PI_year = [str(i) for i in PIData.Year.to_list()]
        except Exception as e:
            logging.debug(f"Error while working with Indicator Data {e}")
            return None
        try:
            index = PI_year.index(year)
            PI_score = []
            for i in code:
                if str(i) in PIData.columns.to_list():
                    PI_score.append(PIData[str(i)].tolist()[index])
                else:
                    # The political instability score from the WGI is not provided for few countries
                    # We assign them a score of 0.5 as most of these countries fall in this range
                    PI_score.append(0.5)
        except Exception as e:
            logging.debug(f"Error while working with Indicator Data {e}")
            return None

        # Version 0.2: Domestic recycling mitigates the supply risk of a raw material. However domestic recycling
        # can be attributed to decrease of imports from a country with low WGI score or high WGI score.
        # In other words, there can be two scenarios where imports are reduced from
        # a riskier country (best case scenario) or imports are reduced from
        # much stable country (worst case scenario). Both the cases are determined by the
        # WGI score, a higher WGI score is for a riskier country while lower for a stable
        # country. The following code intends to manipulate the trade data to incorporate
        # the domestic recycling.

        # Recyclability factor of GeoPolRisk

        # Usually users are supposed to provide an input between 0 and 1
        if recyclingrate >= 1.00001 and recyclingrate < 100:
            recyclingrate = recyclingrate / 100
        elif recyclingrate >= 0 and recyclingrate < 1.00001:
            recyclingrate = recyclingrate
        else:
            logging.debug(
                f"Recycling Rate out of bounds| Recycling Rate : {recyclingrate}"
            )
            recyclingrate = 0

        # The mitigation due to recycling has a redistribution and reduction effect
        # The reduction effect is symbolical because the reduced trade is included in the domestic production
        # Two cases are assumed for the redistribution
        # The function takes in the trade quantity, political instability indicator and sum of the trade quantity
        # This function sorts and reduces the trade quantity based on the scenario and returns the reduced quantity
        def redistribution(quantity, indicator, totQ, reverse):
            try:
                totQ = totQ * recyclingrate
                temp = [(v, i) for i, v in enumerate(indicator)]
                temp.sort(reverse=reverse)
                sortedVal, sortedInd = zip(*temp)
                dump = 0
                for i, n in enumerate(sortedInd):
                    dump += quantity[n]
                    if (totQ - dump) < quantity[sortedInd[i + 1]]:
                        quantity[sortedInd[i + 1]] = quantity[sortedInd[i + 1]] - (
                            totQ - dump
                        )
                        quantity[n] = 0
                        break
                    else:
                        quantity[n] = 0
            except Exception as e:
                logging.debug(f"Error while redistributing the trade {e}")
                quantity = None
            return quantity

        totQ = sum(quantity)
        if scenario == 1:  # Best case scenario
            newquantity = redistribution(quantity, PI_score, totQ, True)
            if newquantity is not None:
                try:
                    zipped_list = zip(newquantity, PI_score)
                    wgiavg = [x * y for (x, y) in zipped_list]
                except Exception as e:
                    logging.debug(f"Error while weighting the trade data {e}")
                    return None
            else:
                logging.debug("Error while redistributing the trade")
                return None
        elif scenario == 2:  # Worst case scenario
            newquantity = redistribution(quantity, PI_score, totQ, False)
            if newquantity is not None:
                try:
                    zipped_list = zip(newquantity, PI_score)
                    wgiavg = [x * y for (x, y) in zipped_list]
                except Exception as e:
                    logging.debug(f"Error while weighting the trade data {e}")
                    return None
            else:
                logging.debug("Error while redistributing the trade")
                return None
        elif scenario == 0:  # No scenario
            try:
                zipped_list = zip(quantity, PI_score)
                wgiavg = [x * y for (x, y) in zipped_list]
            except Exception as e:
                logging.debug(f"Error while weighting the trade data {e}")
                return None

        # After manipulation of the trade data it is multiplied with the WGI
        # score forming the numerator of the second factor of GeoPolRisk (WTA)

        try:
            numerator = sum(wgiavg)
            tradetotal = totQ
        except Exception as e:
            logging.debug(f"Error while summing the weighted trade data{e}")
        return numerator, tradetotal
    else:
        return 0, 0


# The first component of the GeoPolRisk method involved calculating the herfindahl-hirschmann
# index (HHI) and total domestic production required for calculating the second factor (WTA).
def GeoPolRisk(ProductionData, WTAData, Year, AVGPrice):
    newdf = pd.DataFrame(columns=["production"])
    Index = ProductionData[2].index(int(Year))
    HHI = ProductionData[0][Index]
    PQT = ProductionData[1][Index] * 1000  # Converting from Mtonnes to kilograms
    if isinstance(AVGPrice, (int, float)) and WTAData[1] != 0:
        try:
            WTA = WTAData[0] / (WTAData[1] + PQT)
            GeoPolRisk = HHI * WTA
            GeoPolCF = GeoPolRisk * AVGPrice
        except Exception as e:
            logging.debug(f"Error while calculating GeoPolRisk {e}")
            return None
    elif WTAData[1] != 0:
        try:
            WTA = WTAData[0] / (WTAData[1] + PQT)
            GeoPolRisk = HHI * WTA
            GeoPolCF = "NA"
        except Exception as e:
            logging.debug(f"Error while calculating GeoPolRisk {e}")
            return None
    else:
        WTA = 0
        GeoPolRisk = 0
        GeoPolCF = 0
        logging.info("WTA has returned 0")
        return None
    results = [HHI, WTA, GeoPolRisk, GeoPolCF]
    return results



# Copyright (C) 2023 University of Bordeaux, CyVi Group & Anish Koyamparambath
# This file is part of geopolrisk-py library.
#
# geopolrisk-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# geopolrisk-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with geopolrisk-py.  If not, see <https://www.gnu.org/licenses/>.

from .__init__ import (
    instance,
    logging,
    Filename,
    outputdf,
    execute_query,
)
from .core import *
import itertools, pandas as pd, time
from .utils import *
import sys

"""
The module implements the functions of the core module for ease of use.
Provide the list of mineral resources, countries and year to the main function.
Call the regions function to define any custom regions. Call the settradepath
function to define the trade data for specific trade calculation. 
The main function verifies if the above said functions are called or not to 
automatically detect the type of calculation required.
"""


try:
    regionslist, _outputfile = instance.regionslist, instance.exportfile
    _reporter = instance.reporter
    Country = _reporter.Country.to_list()
    _price = instance.price
    _wgi = instance.wgi
    db = _outputfile + "/" + instance.Output
    OutputList = outputdf.outputList
    counter, totalcounter, emptycounter = 0, 0, 0
except Exception as e:
    logging.debug(f"Error with database files or init file {e}")


def tradeagg(resource, year, listofcountries, sheetname=None):
    """
    This function is used to aggregate the trade data from COMTRADE API.
    The trade data is aggregated (summed) which is then
    processed for further calculation.
    """
    # resources is hs code of the commodity
    # New variables to aggregate trade data from COMTRADE API
    newcodelist = []
    newcountrylist = []
    newquantitylist = []
    # TradeData (from core module) fetchs the ISO country code,
    # country names and trade quantity in KG from COMTRADE
    newtradelist = [
        newcodelist,
        newcountrylist,
        newquantitylist,
    ]
    # Variable to fetch production data
    TotalDomesticProduction, TDP = 0, []
    # If the aggregation is not from the specific trade data source,
    # then its not necessary to call the settradepath function and
    # input sheetname as None.
    if sheetname is None:
        for k in listofcountries:
            try:
                TradeData = worldtrade(
                    year=year,
                    country=k,
                    commodity=resource,
                )
            except Exception as e:
                logging.debug(
                    f"""Failed COMTRADE attempt!: resource - {resource}
                            country - {k}, year - {year}, : {e}"""
                )
            # Logging null return of the trade data from the world trade.
            if (
                isinstance(TradeData[0], type(None))
                or len(TradeData) == 0
                or any(v is None for v in TradeData[0])
            ):
                logging.debug(f"Trade of {resource} to {k} returned None for {year}")
                newcodelist.append(0)
                newcountrylist.append(0)
                newquantitylist.append(0)
            else:
                # code to aggregate the data
                TradeData[2] = [0 if v is None else v for v in TradeData[2]]
                for ind, n in enumerate(TradeData[0]):
                    # Aggregation of the trade data
                    # EX: for a region of two countries (A, B) that import from similar set of countries
                    # The aggregation is the sum of the trade quantities for the two countries (A, B)
                    if n not in newcodelist:
                        newcodelist.append(n)
                        newquantitylist.append(TradeData[2][ind])
                        newcountrylist.append(TradeData[1][ind])
                    else:
                        index = newcodelist.index(n)
                        newquantitylist[index] = (
                            newquantitylist[index] + TradeData[2][ind]
                        )

                logging.debug(f"The aggregated trade quantity {newtradelist}")

            # Aggregation of the production of mineral resource data
            xresource, country, _ignore, _ignore2 = convertCodes(resource, k)
            X = ProductionData(xresource, country)
            index = X[2].index(year)
            TotalDomesticProduction += X[1][index]
            logging.info(f"The aggegated production quantity {TotalDomesticProduction}")
    else:
        TradeData = specifictrade(sheetname)
        for k in listofcountries:
            xresource, country, _ignore, _ignore2 = convertCodes(resource, k)
            X = ProductionData(xresource, country)
            index = X[2].index(year)
            TotalDomesticProduction += X[1][index]
        newtradelist = TradeData

    # Calculation of the components of the GeoPolRisk method
    TDP = [0] * len(X[1])
    TDP[X[2].index(year)] = TotalDomesticProduction
    productiondata = [X[0], TDP, X[2]]

    return newtradelist, productiondata


def main_complete(
    resourcelist,
    yearlist,
    countrylist,
    recyclingrate=0.0,
    scenario=0,
    sheetname=None,
    PIindicator=None,
):

    if PIindicator is None:
        PIindicator = _wgi
    # Instantiate counters for logging
    logging.info("Running")
    newregion = []
    if len(regionslist) > 1:
        logging.info("Regional calculation started!")
        newregionlist = []
        # Compare if the region already exists withing the library.
        newregion = [
            i
            for i, x in enumerate(countrylist)
            if str(x) not in Country or str(x) not in [str(i) for i in ISO]
        ]
        logging.info(f"List of new regions {newregion}")

        # Extract new regions from the provided countrylist parameter (argument).
        if len(newregion) > 0:
            for i in newregion:
                newregionlist.append(countrylist[i])
                del countrylist[i]
        if len(newregionlist) > 0:  # Calculate the values for the new region(s)
            for l in newregionlist:
                # New regions can be defined as ISO nomenclature or ISO numeric.
                #  The values must be converted to numeric.newcountrylist
                ncountrylist = regionslist[l]
                logging.info(
                    f"The new region defined {l} and the countries in it are {ncountrylist}"
                )
                _ignore1, _ignore2, _ignore3, Xcountrylist = convertCodes(
                    ["Cobalt"], ncountrylist
                )

                # Iterate the resources, year for COMTRADE query
                for I in itertools.product(resourcelist, yearlist):
                    # LOGIC
                    DBID = create_id(I[0], l, I[1])
                    try:
                        verify = sqlverify(DBID)
                    except Exception as e:
                        logging.debug(f"SQL Verification failed! {e}")
                    if verify is False:
                        time.sleep(2)
                        TradeData, productiondata = tradeagg(I[0], I[1], Xcountrylist)
                        AVGPrice = _price[str(I[1])].tolist()[
                            _price.HS.to_list().index(I[0])
                        ]
                        Y = weightedtrade(
                            str(I[1]),
                            TradeData=TradeData,
                            PIData=PIindicator,
                            scenario=scenario,
                            recyclingrate=recyclingrate,
                        )
                        if Y is not None and productiondata is not None:
                            result = GeoPolRisk(productiondata, Y, str(I[1]), AVGPrice)
                            if result is not None:
                                HHI, WTA, Risk, CF = (
                                    result[0],
                                    result[1],
                                    result[2],
                                    result[3],
                                )
                            else:
                                HHI, WTA, Risk, CF = 0, 0, 0, 0
                        else:
                            HHI, WTA, Risk, CF = 0, 0, 0, 0
                        # Final outputs from the GeoPolRisk calculation
                        logging.info(
                            f"""
                        GeoPolRisk calculation completed successfully! The following
                        values are calculated: {HHI}, {WTA}, {Risk}, {CF} for resource {I[0]}
                        , year {I[1]} to region {l} with recyclingrate {recyclingrate} and 
                        scenario {scenario}
                        """
                        )
                        resource, _ignore1, _ignore2, _ignore3 = convertCodes(I[0], 174)
                        # Record or update data
                        recordData(
                            resource,
                            l,
                            I[1],
                            recyclingrate,
                            scenario,
                            Risk,
                            CF,
                            HHI,
                            WTA,
                            Filename,
                            OutputList,
                        )
                    else:
                        try:
                            sqlstatement = f""" SELECT  geopolrisk, hhi, wta, geopol_cf FROM
                            recordData WHERE id = '{DBID}'
                            """
                            row = execute_query(sqlstatement, db_path=db)
                            OutputList.append(
                                [
                                    str(I[1]),
                                    str(resource),
                                    str(l),
                                    str(recyclingrate),
                                    str(scenario),
                                    str(row[0][0]),
                                    str(row[0][3]),
                                    str(row[0][1]),
                                    str(row[0][2]),
                                ]
                            )
                        except Exception as e:
                            logging.debug(f"Cannot append into list of records {e}")
                        logging.debug(
                            "No transaction has been made. "
                            "Preexisting data has been inserted in output file."
                        )
        else:
            startmain(
                resourcelist,
                yearlist,
                countrylist,
                recyclingrate=0.0,
                scenario=0,
                sheetname=None,
                PIindicator=None,
            )
    else:
        regions()
        startmain(
            resourcelist,
            yearlist,
            countrylist,
            recyclingrate=0.0,
            scenario=0,
            sheetname=None,
            PIindicator=None,
        )

    endmain()


def startmain(
    resourcelist,
    yearlist,
    countrylist,
    recyclingrate=0.0,
    scenario=0,
    sheetname=None,
    PIindicator=None,
):
    if PIindicator is None:
        PIindicator = _wgi
    for I in itertools.product(resourcelist, countrylist, yearlist):
        # LOGIC
        # Need to verify if the data preexists to avoid limited API calls
        resource, country, _ignore, _ignore1 = convertCodes(I[0], I[1])
        DBID = create_id(I[0], I[1], I[2])
        try:
            verify = sqlverify(DBID)
        except Exception as e:
            logging.debug(f"Error with sqlverify or DBID {e}")
        if verify is False:
            # The program has to sleep inorder to avoid conflict in multiple API requests
            time.sleep(2)
            if sheetname is None:
                try:
                    TradeData = worldtrade(
                        year=I[2],
                        country=I[1],
                        commodity=I[0],
                    )
                except Exception as e:
                    logging.debug(f"Error accessing world trade data {e}")
                    counter -= 1
                    break
            else:
                try:
                    TradeData = specifictrade(
                        year=I[2],
                        country=I[1],
                        commodity=I[0],
                    )
                except Exception as e:
                    logging.debug(f"Error accessing specific trade data {e}")
                    break

            try:
                AVGPrice = _price[str(I[2])].tolist()[_price.HS.to_list().index(I[0])]
                X = ProductionData(resource, country)
                Y = weightedtrade(
                    str(I[2]),
                    TradeData=TradeData,
                    PIData=PIindicator,
                    scenario=scenario,
                    recyclingrate=recyclingrate,
                )
                if Y is not None and X is not None:
                    result = GeoPolRisk(X, Y, str(I[2]), AVGPrice)
                    if result is not None:
                        HHI, WTA, Risk, CF = (
                            result[0],
                            result[1],
                            result[2],
                            result[3],
                        )
                    else:
                        HHI, WTA, Risk, CF = 0, 0, 0, 0
                else:
                    HHI, WTA, Risk, CF = 0, 0, 0, 0
            except Exception as e:
                logging.debug(
                    f"Error unpacking the production data/weighted trade data/price data {e}"
                )
                continue

            recordData(
                resource,
                country,
                I[2],
                recyclingrate,
                scenario,
                Risk,
                CF,
                HHI,
                WTA,
                Filename,
                OutputList,
            )
        else:
            try:
                sqlstatement = f""" SELECT  geopolrisk, hhi, wta, geopol_cf FROM
                recordData WHERE id = '{DBID}'
                """
                row = execute_query(sqlstatement, db_path=db)
                OutputList.append(
                    [
                        str(I[1]),
                        str(resource),
                        str(country),
                        str(recyclingrate),
                        str(scenario),
                        str(row[0][0]),
                        str(row[0][3]),
                        str(row[0][1]),
                        str(row[0][2]),
                    ]
                )
            except Exception as e:
                logging.debug(f"Cannot append into list of records {e}")
            logging.debug(
                "No transaction has been made. "
                "Preexisting data has been inserted in output file."
            )


def endmain():
    outputDF = pd.DataFrame(OutputList, columns=outputdf.columns)
    outputDF.to_csv(_outputfile + "/export.csv", index=False, encoding="utf-8")
    logging.info("GeoPolRisk calculation completed successfully!")


def update_cf():
    # Fetch Missing Data
    sqlstatement = (
        "SELECT resource_hscode, year, iso, geopol_cf FROM recordData WHERE "
        " wta = '0';"
    )
    row = execute_query(sqlstatement, db_path=db)
    df = pd.DataFrame(
        row, columns=["HS Code", "Year", "Country Alpha", "Characterziation Factors"]
    )
    logging.debug("Update of database! The shape of df is " + str(df.shape[0]))
    if df.shape[0] > 0:
        Year = [int(i) for i in df.Year.to_list()]
        ISO = [int(i) for i in df["Country Alpha"].tolist()]
        HS = [int(i) for i in df["HS Code"].tolist()]
    else:
        logging.debug("No updates required!")
        return None
    try:
        startmain(HS, Year, ISO, 0, 0)
        endmain()
    except Exception as e:
        logging.debug(f"Error while updating CF! {e}")


# Run this function if there is an update in the missing price data in the average yearly price json.
def updateprice():
    logging.info("Updating the characterization prices | price")
    sqlstatement = (
        "SELECT resource_hscode, year, iso, geopolrisk FROM recordData WHERE "
        " geopol_cf = 'NA' AND recycling_rate ='0' AND scenario ='0';"
    )
    row = execute_query(sqlstatement)
    df = pd.DataFrame(row, columns=["HS Code", "Year", "Country Alpha", "GeoPolRisk"])
    logging.debug("Update of database! The shape of df is " + str(df.shape[0]))
    if df.shape[0] > 0:
        Year = [int(i) for i in df.Year.to_list()]
        ISO = [int(i) for i in df["Country Alpha"].tolist()]
        _HS = [int(i) for i in df["HS Code"].tolist()]
        GPRS = [float(i) for i in df["GeoPolRisk"].tolist()]
    else:
        logging.debug("No updates required!")
        return None
    for i, n in enumerate(_HS):
        AVGPrice = _price[str(Year[i])].tolist()[_price.HS.to_list().index(n)]
        if isinstance(AVGPrice, (int, float)):
            index = ISO.index(ISO[i])
            CF = float(GPRS[index]) * AVGPrice
            sqlstatement = (
                "UPDATE recordData SET geopol_cf= '"
                + str(CF)
                + "', log_ref='"
                + str(Filename)
                + "' WHERE iso = '"
                + str(ISO[i])
                + "' AND resource_hscode= '"
                + str(n)
                + "' AND year = '"
                + str(Year[i])
                + "' AND recycling_rate = '0' AND scenario = '0';"
            )
            norow = execute_query(sqlstatement, db_path=db)
            logging.debug(f"Database update sucessfully! for {n} for {Year[i]}")



# Copyright (C) 2023 University of Bordeaux, CyVi Group & Anish Koyamparambath
# This file is part of geopolrisk-py library.
#
# geopolrisk-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# geopolrisk-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with geopolrisk-py.  If not, see <https://www.gnu.org/licenses/>.

import pandas as pd, json, sqlite3
import comtradeapicall as ctac
from urllib.request import Request, urlopen
from .__init__ import instance, logging, execute_query
from .main import startmain as gprs_comtrade
from .main import endmain
from .utils import convertCodes


# Define Paths
tradepath = None
_production, _reporter = instance.production, instance.reporter
regionslist, _outputfile = instance.regionslist, instance.exportfile
_price = instance.price
db = _outputfile + "/" + instance.Output
# Extract list of all data
HS = _price.HS.to_list()
HS = [int(float(x)) for x in HS]
Resource = _price.Resource.to_list()
Country = _reporter.Country.to_list()
ISO = _reporter.ISO.to_list()
ISO = [int(x) for x in ISO]


# from .Exceptions.warningsgprs import *
import itertools, sqlite3, pandas as pd, sys
from difflib import get_close_matches


"""
A simple guided step by step method for new users. This method can be called 
to any console based application if needed.
"""


def matchwords(word, List):
    word = word.lower()
    List = [x.lower() for x in List]
    if word in List:
        index = List.index(word)
        return index
    else:
        return -1


def getthequestions(compareList, *args):
    A = args[0]
    B = args[1]
    print(A)
    _exit = True
    finalvalue = False
    while _exit:
        inputvalue = input("Enter the value: ")
        comparelist = list(map(lambda x: x.lower(), compareList))
        closevalues = get_close_matches(inputvalue.lower(), comparelist)
        cl_index = matchwords(inputvalue, comparelist)
        if cl_index > -1:
            finalvalue = compareList[cl_index]
            _exit = False
        elif inputvalue.lower() == "exit":
            sys.exit(1)
        elif len(closevalues) == 1:
            cl_index = matchwords(closevalues[0], comparelist)
            if cl_index > -1:
                finalvalue = compareList[cl_index]
                _exit = False
            else:
                print("Please try again.")
        elif len(closevalues) > 1:
            print("Input did not match exactly a value in our list.")
            print("Is the value you intented in the list below?")
            print(closevalues)
            print("If yes, input the position of the value as 1 ,2 ,3,...")
            print("If not, type 'no'.")
            ni_value = input("Inpute the position : ")
            try:
                ni_value = int(ni_value) - 1
                _exit = False
            except:
                _exit = True
                break

            cl_index = matchwords(closevalues[ni_value], comparelist)
            if cl_index > -1:
                finalvalue = compareList[cl_index]
                _exit = False
            else:
                print("")
                _exit = True
                break
        else:
            print(B)

    return finalvalue


def guided():
    start = True
    intro = """
    Welcome to the geopolrisk-py guided assessment. Using this function, a user can
    assess the supply risk of importing a resource to a country. Questions are asked
    at each step of the assessment to which the user must type the response. At any 
    given point, the user can exit the assessment by typing 'exit'. The results of 
    the assessment are downloaded to the output folder.
    """
    print(intro)
    logging.info("Assessment type: Guided")
    while start:
        A = "Enter the name of the resource : "
        B = "The entered resource is not found! Please try again."
        resource = [
            getthequestions(Resource, A, B),
        ]
        logging.debug("Guided Assessment| Input Resource: {}".format(resource))
        A = "Enter the name of the country of asssessment : "
        B = "The entered country is not found! Please try again."
        country = [
            getthequestions(Country, A, B),
        ]
        logging.debug("Guided Assessment| Input Country: {}".format(country))
        _exit = True
        while _exit:
            period = input("Enter the year of assessment :")
            try:
                if int(period) in range(2000, 2020):
                    period = int(period)
                    logging.debug("Guided Assessment| Input Year: {}".format(period))
                    _exit = False
                else:
                    print("Provide enter an year between 2000 and 2020")
                    _exit = True
            except Exception as e:
                logging.debug(e)
                if period.lower() == "exit":
                    sys.exit()
                break

        _exit = True
        while _exit:
            print(
                "Enter the rate of domestic recycling input of the raw material (in fractions) or type exit to exit"
            )
            recyclingrate = input("Enter the recycling input rate: ")
            try:
                if float(recyclingrate) in range(0, 101):
                    recyclingrate = float(recyclingrate) / 100
                    logging.debug(
                        "Guided Assessment| Input Recycling Rate: {}".format(
                            recyclingrate
                        )
                    )
                    _exit = False
                else:
                    print("Provide enter a value between 0 to 100.")
                    _exit = True
            except Exception as e:
                logging.debug(e)
                if recyclingrate.lower() == "exit":
                    sys.exit()
                break

        _exit = True
        while _exit:
            print("Select 0 for no recycling scenario (recyclingrate is set to zero)")
            print(
                "Select 0 for assuming a best case scenario where the imports from a riskier country is reduced"
            )
            print(
                "Select 1 for assuming a worst case scenario where the imports from a stable country is reduced"
            )
            scenario = input("Select 1 or 0: ")
            try:
                if int(scenario) in [0, 1]:
                    scenario = int(scenario)
                    logging.debug(
                        "Guided Assessment| Input Scenario: {}".format(scenario)
                    )
                    _exit = False
                else:
                    print("Provide enter either 0 or 1.")
                    _exit = True
            except Exception as e:
                logging.debug(e)
                if scenario.lower() == "exit":
                    sys.exit()
                break
        _ignore, _ignore2, resource, country = convertCodes(resource, country)
        gprs_comtrade(resource, [period], country, recyclingrate, scenario)
        endmain()

        print("Assessment completed! The results are downloaded to the output folder")
        print("In case of error in the output file, refer to logs.")



# Copyright (C) 2023 University of Bordeaux, CyVi Group & Anish Koyamparambath
# This file is part of geopolrisk-py library.
#
# geopolrisk-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# geopolrisk-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with geopolrisk-py.  If not, see <https://www.gnu.org/licenses/>.

import sqlite3, pandas as pd, getpass, logging, os
from datetime import datetime
from pathlib import Path

# from .Exceptions.warningsgprs import *

logging = logging
__all__ = ["core", "operations", "console", "gprsplots", "utils", "tests", "main"]
__author__ = "Anish Koyamparambath <CyVi- University of Bordeaux>"
__status__ = "release"
__version__ = "1"
__data__ = "10 March 2023"

hard_dependencies = ("pandas", "scipy", "matplotlib")
missing_dependencies = []
for dependency in hard_dependencies:
    try:
        __import__(dependency)
    except ImportError as e:
        missing_dependencies.append(f"{dependency}: {e}")

if missing_dependencies:
    raise ImportError(
        "Unable to import required dependencies:\n" + "\n".join(missing_dependencies)
    )
del hard_dependencies, dependency, missing_dependencies


# Function to use sqlite3
def execute_query(query, db_path=""):
    # Connect to the database
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    # Determine if the query is a SELECT or INSERT query
    is_select_query = query.strip().lower().startswith("select")

    # Execute the query and fetch the results (if it is a SELECT query)
    if is_select_query:
        cursor.execute(query)
        results = cursor.fetchall()
    else:
        cursor.execute(query)
        results = None

    # Commit the query to the database
    conn.commit()

    # Close the database connection
    conn.close()

    # Return the results (if it is a SELECT query)
    if is_select_query:
        return results


class database:
    # Global Variables
    Output = "Datarecords.db"
    Database = "library.db"
    OutputFolder = "Documents/geopolrisk"
    LogFolder = OutputFolder + "/logs"
    CFDatabase = OutputFolder + "/output"
    DBFolder = OutputFolder + "/databases"

    # Create directories
    try:
        directory = getpass.getuser()
        if not os.path.exists(os.path.join(Path.home(), OutputFolder)):
            os.makedirs(os.path.join(Path.home(), OutputFolder))

        logfile = os.path.join(Path.home(), LogFolder)
        if not os.path.exists(logfile):
            os.makedirs(logfile)

        dbFolder = os.path.join(Path.home(), DBFolder)
        if not os.path.exists(dbFolder):
            os.makedirs(dbFolder)

        exportfile = os.path.join(Path.home(), CFDatabase)
        if not os.path.exists(exportfile):
            os.makedirs(exportfile)
    except Exception as e:
        print(f"Unable to create directories {e}")
        raise FileNotFoundError

    # Verify that the file exists else input the correct file path
    if not os.path.isfile(os.path.join(dbFolder, Database)):
        folder_path = input("The file doesn't exist. Please enter a folder path: ")

    # Verify if the database contains the required tables
    Tables = [
        "commodityHS",
        "Country_ISO",
        "Aluminium",
        "Antimony",
        "Asbestos",
        "Barytes",
        "Bismuth",
        "Cadmium",
        "Chromium",
        "Coal",
        "Cobalt",
        "Copper",
        "Gold",
        "Graphite",
        "Iron",
        "Lead",
        "Lithium",
        "Magnesite",
        "Magnesium",
        "Manganese",
        "Mercury",
        "Molybdenum",
        "Nickel",
        "Crude_Oil",
        "REE",
        "Silver",
        "Tin",
        "Titanium",
        "Tungsten",
        "Uranium",
        "Zinc",
        "Zirconium",
        "NG",
        "WGI",
        "Price",
    ]

    db = dbFolder + "/" + Database
    # Check if the database exists and fetch the required tables
    try:
        result = execute_query(
            "SELECT name FROM sqlite_master WHERE type='table';", db_path=db
        )
        table_names = [row[0] for row in result]
    except Exception as e:
        print(f"Unable to verify if the database contains the required tables {e}")
        raise FileNotFoundError

    # check if all the tables in the list are present in the database
    missingTables = []
    for table_name in Tables:
        if table_name not in table_names:
            missingTables.append(table_name)
        else:
            pass

    # If there are missing tables, raise an error
    if len(missingTables) > 0:
        print(f"The following tables are missing from the database: {missingTables}")
    else:
        pass

    # Function to extract the tables into a dictionary
    def extract_tables_to_df(db_path, table_names):
        # Create a dictionary to store the extracted tables
        tables = {}

        # Connect to the database
        conn = sqlite3.connect(db_path)

        # Loop through the table names and extract each table as a DataFrame
        for table_name in table_names:
            query = f"SELECT * FROM {table_name}"
            table_df = pd.read_sql_query(query, conn)

            # Add the table DataFrame to the tables dictionary with the table name as the key
            tables[table_name] = table_df

        # Close the database connection
        conn.close()

        # Return the tables dictionary
        return tables

    # verify if output database exists else create it
    try:
        sqlstatement = """CREATE TABLE IF NOT EXISTS "recordData" (
        	"index"	INTEGER,
            "id" TEXT NOT NULL,
        	"country"	TEXT,
        	"resource"	TEXT,
        	"year"	INTEGER,
        	"recycling_rate"	REAL,
        	"scenario"	REAL,
        	"geopolrisk"	REAL,
        	"hhi"	REAL,
        	"wta"	REAL,
        	"geopol_cf"	REAL,
        	"resource_hscode"	INTEGER,
        	"iso"	INTEGER,
            "log_ref" TEXT,
        	PRIMARY KEY("index")
        );"""
        execute_query(sqlstatement, db_path=exportfile + "/" + Output)
    except Exception as e:
        print(f"Could not create the output database {e}")

    # Extract the tables into a dictionary
    try:
        tables = extract_tables_to_df(db, Tables)
    except Exception as e:
        print(f"Could not extract the tables {e}")

    production = tables
    reporter = tables["Country_ISO"]
    price = tables["Price"]
    commodity = tables["commodityHS"]
    wgi = tables["WGI"]
    regionslist = {}
    regionslist["EU"] = [
        "Austria",
        "Belgium",
        "Belgium-Luxembourg",
        "Bulgaria",
        "Croatia",
        "Czechia",
        "Czechoslovakia",
        "Denmark",
        "Estonia",
        "Finland",
        "France",
        "Fmr Dem. Rep. of Germany",
        "Fmr Fed. Rep. of Germany",
        "Germany",
        "Greece",
        "Hungary",
        "Ireland",
        "Italy",
        "Latvia",
        "Lithuania",
        "Luxembourg",
        "Malta",
        "Netherlands",
        "Poland",
        "Portugal",
        "Romania",
        "Slovakia",
        "Slovenia",
        "Spain",
        "Sweden",
    ]


instance = database()


class outputDF:
    columns = [
        "Year",
        "Resource",
        "Country",
        "Recycling Rate",
        "Recycling Scenario",
        "Risk",
        "GeoPolRisk Characterization Factor",
        "HHI",
        "Weighted Trade AVerage",
    ]
    outputList = []


outputdf = outputDF()
# Test fail variables
LOGFAIL, DBIMPORTFAIL = False, False

# Create a log file for init function

Filename = "Log_File_{:%Y-%m-%d(%H-%M-%S)}.log".format(datetime.now())

log_level = logging.DEBUG
try:
    logging.basicConfig(
        level=log_level,
        format="""%(asctime)s | %(levelname)s | %(threadName)-10s |
          %(filename)s:%(lineno)s - %(funcName)20s() |
            %(message)s""",
        filename=instance.logfile + "/" + Filename,
        filemode="w",
    )
except:
    # it is imperative that the log file work before running the main code.
    LOGFAIL = True
    print("Cannot create log file!")



# Copyright (C) 2023 University of Bordeaux, CyVi Group & Anish Koyamparambath
# This file is part of geopolrisk-py library.
#
# geopolrisk-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# geopolrisk-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with geopolrisk-py.  If not, see <https://www.gnu.org/licenses/>.

import unittest
from .defaults import *
from geopolrisk.assessment.core import regions
from geopolrisk.assessment.main import startmain, endmain, main_complete
from geopolrisk.assessment.__init__ import instance

_wgi = instance.wgi
# testcase to test the gprs_comtrade function
Resources = random_sampler(ListofMetals, 2)
Countries = random_sampler(ListofCountries, 2)
Year = random_sampler(ListofYears, 2)
listofcountryname, listofresourcename = [], []
for i, n in enumerate(Resources):
    listofcountryname.append(ListofCountryName[ListofCountries.index(Countries[i])])
    listofresourcename.append(ListofMetalName[ListofMetals.index(n)])

class Testmainfunctions(unittest.TestCase):
    def test_aggregion(self):
        NewRegion = {"Custom Region" : listofcountryname }
        xx = regions(NewRegion)
        try:
            main_complete(Resources,Year,["Custom Region"],0,0,None,_wgi)
            assert True
        except Exception as e:
            print(e)
            assert False

    def test_mainfunction(self):
        try:
            startmain(
            Resources,
            Year,
            Countries,
            recyclingrate=0.0,
            scenario=0,
            sheetname=None,
            PIindicator=None,
            )
            endmain()
        except Exception as e:
            print(e)
            assert False



# Copyright (C) 2023 University of Bordeaux, CyVi Group & Anish Koyamparambath
# This file is part of geopolrisk-py library.
#
# geopolrisk-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# geopolrisk-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with geopolrisk-py.  If not, see <https://www.gnu.org/licenses/>.
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys
import subprocess


class RunTests(TestCommand):
    """
    Custom test command that runs the tests defined in the 'tests.py' file.
    """
    def run_tests(self):
        # Run the tests using the 'unittest' module
        test_file = 'geopolrisk/tests.py'
        test_result = subprocess.run([sys.executable, test_file], stdout=subprocess.PIPE)
        
        # Display the test results to the user
        print(test_result.stdout.decode('utf-8'))

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='geopolrisk-py',
    version='1.1',
    author="Anish Koyamparambath",
    author_email="anish.koyamparambath@u-bordeaux.fr",
    description="A complete library for calculation of the GeoPolRisk method",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/akoyamp/geopolrisk-py",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.7",
    install_requires=[
        'pandas',
        'matplotlib',
        'scipy',
        'comtradeapicall',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    license="GPL-3.0-or-later",
    cmdclass={
        'test': RunTests,
    },
)






# Copyright (C) 2023 University of Bordeaux, CyVi Group & Anish Koyamparambath
# This file is part of geopolrisk-py library.
#
# geopolrisk-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# geopolrisk-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with geopolrisk-py.  If not, see <https://www.gnu.org/licenses/>.

from matplotlib import rcParams

rcParams["font.family"] = "sans-serif"
rcParams["font.sans-serif"] = ["Tahoma"]
import pandas as pd, numpy as np, matplotlib.pyplot as plt
from scipy.interpolate import make_interp_spline, BSpline
from matplotlib.ticker import MaxNLocator
from textwrap import wrap
from .__init__ import (
    instance,
    logging,
)
_outputfile = instance.exportfile
logging.info("Accessed gprsplots module.")

# Read output file
global csvfile
csvfile = "/export.csv"
try:
    data = pd.read_csv(_outputfile + csvfile)
except Exception as e:
    logging.debug(e)

try:
    data.columns = [
        "Year",
        "Resource",
        "Country",
        "RR",
        "RRScenario",
        "GeoPolRisk",
        "CF",
        "HHI",
        "WTA",
    ]
except Exception as e:
    logging.debug(e)


# Splining function
def splining(x, y, degree):
    try:
        x = np.array(x)
        y = np.array(y)
        xnew = np.linspace(x.min(), x.max(), 1000)
        spl = make_interp_spline(x, y, k=degree)  # type: BSpline
        smooth = spl(xnew)
    except Exception as e:
        logging.debug(e)
    return xnew, smooth


def trendplot(country, period, resources):
    fig = plt.figure(figsize=(7, 5), dpi=200, facecolor="#f4f6fa", edgecolor="k")
    gs = fig.add_gridspec(1, 1)
    fig.suptitle("Evolution of the supply risk of selected resources", fontsize=14)
    ax = fig.add_subplot(gs[0, 0])
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    plt.yticks(np.arange(0, 1, 0.1))

    for i in resources:
        GPRS, year = [], []
        rm_temp = 0
        for j in period:
            try:
                year.append(int(j))
            except Exception as e:
                logging.debug(e)
            try:
                GPRS.append(
                    data.loc[
                        (data["Resource"] == i)
                        & (data["Country"] == country[0])
                        & (data["Year"] == j)
                        & (data["RRScenario"] == 0),
                        "GeoPolRisk",
                    ].values[0]
                )
            except Exception as e:
                logging.debug(e)
        rm = max(GPRS)
        rm_temp = rm if rm > rm_temp else rm_temp
        try:
            x, y = splining(year, GPRS, 2)
            plt.plot(x, y, label=i)
        except Exception as e:
            logging.debug(e)

    plt.xlabel("Year")
    plt.ylabel("GPRS")
    ax.set_ylim([0, rm_temp + 0.1])
    handles, labels = ax.get_legend_handles_labels()
    fig.legend(handles, labels, loc="center right")
    fig.tight_layout()
    fig.subplots_adjust(top=0.90, right=0.80, left=0.15)
    return fig


def indplot(Country, Year, resource, scenario):
    Year = int(Year[0])
    switch_dict = {"Resource": resource, "Country": Country}
    if len(resource) > 1 and len(Country) == 1:
        switch = switch_dict["Resource"]
        catch = switch_dict["Country"]
    else:
        switch = switch_dict["Country"]
        catch = switch_dict["Resource"]

    fig = plt.figure(figsize=(7, 5), dpi=200, facecolor="#f4f6fa", edgecolor="k")
    gs = fig.add_gridspec(1, 1)
    fig.suptitle("Comparison of supply risk of resources", fontsize=14)
    ax = fig.add_subplot(gs[0, 0])
    plt.yticks(np.arange(0, 1, 0.1))
    for i in switch:
        try:
            if scenario == 0:
                _Year = (
                    data.loc[
                        (
                            data[
                                list(switch_dict.keys())[
                                    list(switch_dict.values()).index(switch)
                                ]
                            ]
                            == i
                        )
                        & (
                            data[
                                list(switch_dict.keys())[
                                    list(switch_dict.values()).index(catch)
                                ]
                            ]
                            == switch_dict[
                                list(switch_dict.keys())[
                                    list(switch_dict.values()).index(catch)
                                ]
                            ][0]
                        ),
                        "Year",
                    ]
                    .values.tolist()
                    .index(Year)
                )
                trend = data.loc[
                    (
                        data[
                            list(switch_dict.keys())[
                                list(switch_dict.values()).index(switch)
                            ]
                        ]
                        == i
                    )
                    & (
                        data[
                            list(switch_dict.keys())[
                                list(switch_dict.values()).index(catch)
                            ]
                        ]
                        == switch_dict[
                            list(switch_dict.keys())[
                                list(switch_dict.values()).index(catch)
                            ]
                        ][0]
                    ),
                    "GeoPolRisk",
                ].values[_Year]
                plt.bar([i], trend, label=catch[0])

            else:
                # Best Case Scenario
                _Year = (
                    data.loc[
                        (
                            data[
                                list(switch_dict.keys())[
                                    list(switch_dict.values()).index(switch)
                                ]
                            ]
                            == i
                        )
                        & (
                            data[
                                list(switch_dict.keys())[
                                    list(switch_dict.values()).index(catch)
                                ]
                            ]
                            == switch_dict[
                                list(switch_dict.keys())[
                                    list(switch_dict.values()).index(catch)
                                ]
                            ][0]
                        )
                        & (data["RRScenario"] == 2),
                        "Year",
                    ]
                    .values.tolist()
                    .index(Year)
                )
                trend = data.loc[
                    (
                        data[
                            list(switch_dict.keys())[
                                list(switch_dict.values()).index(switch)
                            ]
                        ]
                        == i
                    )
                    & (
                        data[
                            list(switch_dict.keys())[
                                list(switch_dict.values()).index(catch)
                            ]
                        ]
                        == switch_dict[
                            list(switch_dict.keys())[
                                list(switch_dict.values()).index(catch)
                            ]
                        ][0]
                    )
                    & (data["RRScenario"] == 2),
                    "GeoPolRisk",
                ].values[_Year]
                plt.bar([i], trend, label=catch[0])
                # Worst Case Scenario
                _Year = (
                    data.loc[
                        (
                            data[
                                list(switch_dict.keys())[
                                    list(switch_dict.values()).index(switch)
                                ]
                            ]
                            == i
                        )
                        & (
                            data[
                                list(switch_dict.keys())[
                                    list(switch_dict.values()).index(catch)
                                ]
                            ]
                            == switch_dict[
                                list(switch_dict.keys())[
                                    list(switch_dict.values()).index(catch)
                                ]
                            ][0]
                        )
                        & (data["RRScenario"] == 1),
                        "Year",
                    ]
                    .values.tolist()
                    .index(Year)
                )
                trend = data.loc[
                    (
                        data[
                            list(switch_dict.keys())[
                                list(switch_dict.values()).index(switch)
                            ]
                        ]
                        == i
                    )
                    & (
                        data[
                            list(switch_dict.keys())[
                                list(switch_dict.values()).index(catch)
                            ]
                        ]
                        == switch_dict[
                            list(switch_dict.keys())[
                                list(switch_dict.values()).index(catch)
                            ]
                        ][0]
                    )
                    & (data["RRScenario"] == 1),
                    "GeoPolRisk",
                ].values[_Year]
                plt.bar([i], trend, label=catch[0])
        except Exception as e:
            logging.debug(e)

    plt.ylabel("GPRS")
    ax.set_ylim([0.00, 0.10])
    handles, labels = ax.get_legend_handles_labels()
    labels = ["\n".join(wrap(l, 10)) for l in labels]
    fig.legend(
        handles,
        labels,
        loc="center right",
    )
    fig.tight_layout()
    fig.subplots_adjust(top=0.90, right=0.75, left=0.15)
    return fig


def compareplot(Country, Year, resource, scenario):
    if len(Year) == 1 and scenario == 0:
        Year = int(Year[0])
        fig = plt.figure(figsize=(7, 5), dpi=200, facecolor="#f4f6fa", edgecolor="k")
        gs = fig.add_gridspec(1, 1)
        fig.suptitle("Comparison of supply risk of resources", fontsize=14)
        ax = fig.add_subplot(gs[0, 0])
        plt.yticks(np.arange(0, 1, 0.1))
        columns = ["Country"]
        columns.extend(resource)
        allrows = []
        for i in Country:
            GPRS, HHI, WTA = [], [], []
            dfrow = [i]
            for j in resource:
                GPRS.append(
                    data.loc[
                        (data["Resource"] == j)
                        & (data["Country"] == i)
                        & (data["Year"] == Year)
                        & (data["RRScenario"] == 0),
                        "GeoPolRisk",
                    ].values[0]
                )
                HHI.append(
                    data.loc[
                        (data["Resource"] == j)
                        & (data["Country"] == i)
                        & (data["Year"] == Year)
                        & (data["RRScenario"] == 0),
                        "HHI",
                    ].values[0]
                )
                WTA.append(
                    data.loc[
                        (data["Resource"] == j)
                        & (data["Country"] == i)
                        & (data["Year"] == Year)
                        & (data["RRScenario"] == 0),
                        "WTA",
                    ].values[0]
                )
            total = sum(GPRS)
            GPRS = [(i / total) * 100 for i in GPRS]
            dfrow.extend(GPRS)
            allrows.append(dfrow)
        newpd = pd.DataFrame(allrows, columns=columns)
        newpd.plot(
            x="Country",
            kind="bar",
            stacked=True,
            title="Percentage Stacked Bar Graph",
            mark_right=True,
        )
        plt.ylabel("GPRS")
        plt.legend(loc="center left", bbox_to_anchor=(1, 0.5))
        ax.set_ylim([0.00, 100])
        fig.tight_layout()
        fig.subplots_adjust(top=0.90, right=0.75, left=0.15)
        return fig
